#!/usr/bin/python
# This script implements the constrained least mean square algorithm proposed
# by Chiun Lin Lim. 
# The basic idea of the algorithm is that probe traffic gap variance is minimal
# if the sum of cross traffic rate and probe traffic rate is at the link capacity.
# This has been confirmed empirically with probe traffic + bursty cross traffic.

#
# Algorithm:
# Assume we have a 10Gbps Ethernet link with 1Gbps bursty cross traffic.
# We probe the cross traffic with probe traffic between 8.0 Gbps to 9.4 Gbps, with
# a step of 0.1Gbps, with packet streams of 20 packets. We send 40 packet streams 
# at each probe rate.
#
# The result of gap variance, is an array of 15 numbers. [x, y], x be the probe 
# rate, y be the inter-packet gap variance at that probe rate. 
#
# y is taking the variance of all inter-packet gap within the same stream and 
# average over multiple probe packet streams. 
# 
# If we plot x-y, the graph will have a minimum around 9.0Gbps. To extrapolate 
# the exact bandwidth, we use chiun lin's contrained least mean square algorithm 
# to compute the estimation.

# Input: 
# List of inter-packet gap variance computed after each probe stream.
#
# Output:
# Probe rate at which the variance has a minimum, note, this is extrapolated value.
import sys, os, getopt, math, argparse, string, pickle
import numpy as np
import scipy as sp 

def multiply(x, y):
    return x * y

def square(x):
    return x * x

def sum(x, y):
    return x + y

def object_f(k, b , x, y):
    if len(x) != len(y):
        return 
    result = 0
    for i in range(len(x)):
#        print k, x[i], b, y[i]
        result += square(k * x[i] + b - y[i])
    return result

def lms(y_all):
    n = len(y_all)
    x_all = np.arange(0.1, 9.7, (9.7-0.1)/n)
    xy_all = np.multiply(x_all, y_all)

    r = {}
    rr = []
    for l in range(1, n-1):
        x_lower = x_all[:l+1]
        x_upper = x_all[l+1:]
        sqr_l = map(square, x_lower)
        sqr_u = map(square, x_upper)
        a = np.matrix([[2 * reduce(sum, sqr_l), 0, 2 * reduce(sum, x_lower), 0, x_all[l]],
                       [0, 2 * reduce(sum, sqr_u), 0 , 2* reduce(sum, x_upper), -x_all[l]],
                       [2 * reduce(sum, x_lower), 0, 2 * (l+1), 0, 1],
                       [0, 2 * reduce(sum, x_upper), 0, 2 * (n - l - 1), -1],
                       [x_all[l], -x_all[l], 1, -1 , 0]], dtype=object)
        w = np.matrix([2 * reduce(sum, xy_all[:l+1]),
                       2 * reduce(sum, xy_all[l+1:]),
                       2 * reduce(sum, y_all[:l+1]),
                       2 * reduce(sum, y_all[l+1:]),
                       0])
        v = a.I * w.T
        k1 = v[0]
        k2 = v[1]
        b1 = v[2]
        b2 = v[3]
        lam = v[4]

        target = object_f(k1, b1, x_all[:l+1], y_all[:l+1]) + object_f(k2, b2, x_all[l+1:], y_all[l+1:])
        rr.append(target)
        r[target] = x_all[l]
    return r[min(rr)]
def main():
    parser = argparse.ArgumentParser(description='Analyzing captured data')
    parser.add_argument('-i', '--input', type=str,
            help='input file', required=True)
    parser.add_argument('-v', '--verbose', action='store_true', default=False)
    
    args = parser.parse_args()

    if args.verbose:
        print args

    with open(args.input) as f:
        content = f.readlines()

    y_all = []
    for line in content:
        _line = line.split()
        y_all.append(float(_line[1]))
    y_all = np.asarray(y_all)
    print lms(y_all)

if __name__ == '__main__':
    main()
